import sys


def set_basic_handler(filename: str):
    return [
        {
            "sink": sys.stdout,
            "format": "{message}",
            "level": "ERROR",
        },
        {
            "sink": f"logs/{filename}.log",
            "format": "{message}",
            "level": "WARNING",
            "serialize": True,
        },
    ]


def set_info_handler(filename: str):
    return [
        {
            "sink": sys.stdout,
            "format": "{message}",
            "level": "DEBUG",
        },
        {
            "sink": f"logs/{filename}.log",
            "format": "{message}",
            "level": "INFO",
            "serialize": True,
        },
    ]



import json

from corpus_base import DecisionRow
from sqlpyd import Connection

from .codifications import CodeRow, CodeStatuteEvent, Codification
from .documents import DocRow, Document
from .inclusions import Inclusion
from .resources import get_authored_object
from .statutes import Statute, StatuteFoundInUnit, StatuteRow


def build_x_tables(c: Connection) -> Connection:
    Statute.make_tables(c)
    Codification.make_tables(c)
    Document.make_tables(c)
    Inclusion.make_tables(c)
    return c


def _decode(obj: dict, key: str) -> dict:
    """Convert a string in the `obj` dictionary to a deserialized object value."""
    obj[key] = json.loads(obj.pop(key, []))
    return obj


def get_decision(c: Connection, id: str) -> dict:
    obj = get_authored_object(c, DecisionRow.__tablename__, id)
    obj |= Inclusion.get_base_data(c, id)
    obj |= Inclusion.list_opinions_of_decision(c, id)
    obj = _decode(obj, "author_list")
    obj = _decode(obj, "opinions_list")
    for opinion in obj["opinions_list"]:
        opinion = _decode(opinion, "statutes")
        opinion = _decode(opinion, "unmatched_statutes")
        opinion = _decode(opinion, "decisions")
        opinion = _decode(opinion, "unmatched_decisions")
    return obj


def get_codification(c: Connection, id: str) -> dict:
    obj = get_authored_object(c, CodeRow.__tablename__, id)
    obj |= CodeRow.get_base_data(c, id)
    obj |= CodeStatuteEvent.list_event_statutes(c, id)
    obj = _decode(obj, "statute_titles")
    obj = _decode(obj, "units")
    obj = _decode(obj, "author_list")
    obj = _decode(obj, "event_statute_affectors")
    return obj


def get_statute(c: Connection, id: str) -> dict:
    obj = get_authored_object(c, StatuteRow.__tablename__, id)
    obj |= StatuteRow.get_base_data(c, id)
    obj |= StatuteFoundInUnit.list_affected_statutes(c, id)
    obj |= StatuteFoundInUnit.list_affector_statutes(c, id)
    obj = _decode(obj, "statute_titles")
    obj = _decode(obj, "code_titles")
    obj = _decode(obj, "units")
    obj = _decode(obj, "author_list")
    obj = _decode(obj, "affected_statutes_list")
    obj = _decode(obj, "affector_statutes_list")
    return obj


def get_document(c: Connection, id: str) -> dict:
    obj = get_authored_object(c, DocRow.__tablename__, id)
    # obj = _decode(obj, "units"), needs to be implemented via .get_base_data()
    obj = _decode(obj, "author_list")
    return obj



import abc
import json
from pathlib import Path
from typing import Any

from corpus_pax import Individual
from jinja2 import Environment, PackageLoader, select_autoescape
from loguru import logger
from pydantic import BaseModel, EmailStr
from sqlpyd import Connection

BASE = "code/corpus"

STATUTE_PATH = Path().home().joinpath(f"{BASE}/statutes")

CODIFICATION_PATH = Path().home().joinpath(f"{BASE}/codifications")

DOCUMENT_PATH = Path().home().joinpath(f"{BASE}/documents")

INCLUSION_FILE = "inclusions.yaml"


corpus_sqlenv = Environment(
    loader=PackageLoader(package_name="corpus_x", package_path="sql"),
    autoescape=select_autoescape(),
)


class Integrator(BaseModel, abc.ABC):
    """
    1. Tables need to be created, see `cls.make_tables()`
    2. With `cls.add_rows()`, the content of the table structures are sourced from a local repository processed by `cls.create_obj` functions.
    3. The Integrator instance is a collection of @relations.
    3. Each instance can be added to the tables. See `self.add_to_database()`
    """

    id: str = NotImplemented
    emails: list[EmailStr] = NotImplemented
    meta: Any = NotImplemented
    tree: list[Any] = NotImplemented
    unit_fts: list[Any] = NotImplemented

    @classmethod
    @abc.abstractmethod
    def make_tables(cls, c: Connection) -> None:
        """Common process for creatng the tables associated with the concrete class."""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def add_rows(cls, c: Connection) -> None:
        """Common process for creating objects from the source files for these to become prospective rows to the tables created."""
        raise NotImplementedError

    @abc.abstractmethod
    def add_to_database(self, c: Connection) -> str | None:
        """Each entry of the concrete class is an instance of a pydantic BaseModel, this implies prior validation, and thus can now be added to the tables created in `cls.make_tables()`"""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def from_page(cls, file_path: Path) -> None:
        """The `file_path` expects an appropriate .yaml file containing the metadata. The data will be processed into an interim 'page' that will eventually build an instance of the concrete class."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def relations(cls):
        """Helper property to associate TableConfigured models to their instantiated values in preparation for database insertion."""
        raise NotImplementedError

    def insert_objects(
        self,
        c: Connection,
        obj: Any,
        correlations: list[tuple[Any, Any]],
    ) -> str:
        """The use of the concrete class' `insert_objects()` function implies that an `Individual` table already exists.

        The `obj` is a subclass of `TableConfig`. Since we're already aware of the `id` of the `obj`, we can also use this same id to create the author of the object as well as the correlated entities.

        Each correlated entity must also be a subclass of `TableConfig`.
        """
        record = self.meta.dict(exclude={"emails"})
        c.add_record(obj, record)

        for email in self.emails:
            c.table(obj).update(self.id).m2m(
                other_table=c.table(Individual),
                lookup={"email": email},
                pk="id",
            )

        for related in correlations:
            c.add_cleaned_records(
                related[0],  # the related model which must be
                related[1],  # the instance of the object
            )

        return self.id

    @classmethod
    def create_obj(cls, c: Connection, file_path: Path) -> str | None:
        """Should return `id` of object created, if created,"""
        if obj := cls.from_page(file_path):
            if idx := obj.add_to_database(c):
                return idx
            else:
                logger.error(f"No db entry; see {obj.id}")
        else:
            logger.error(f"No entry created; see {file_path};")
        return None


def sql_get_detail(generic_tbl_name: str, generic_id: str) -> str:
    return corpus_sqlenv.get_template("base/get_detail.sql").render(
        generic_tbl=generic_tbl_name,
        target_id=generic_id,
    )


def sql_get_authors(generic_tbl_name: str, generic_id: str) -> str:
    """Produce the SQL query string necessary to get the authors from the Individual table based on the `generic_tbl_name`'s target `generic_id`.

    Each generic_tbl_name will be sourced from either: DecisionRow, CodeRow, DocRow, StatuteRow. Each of these tables are associated with the Individual table. The result looks something like this:

    >>> from .statutes import StatuteRow
    >>> sql = sql_get_authors(StatuteRow.__tablename__, "ra-386-june-18-1949")
    >>> type(sql)
    <class 'str'>

    See sqlite_utils which creates m2m object tables after sorting the tables alphabetically.
    """
    l = [generic_tbl_name, Individual.__tablename__]
    template = corpus_sqlenv.get_template("base/get_author_ids.sql")
    return template.render(
        generic_tbl="_".join(sorted(l)),  #
        col_generic_obj="_".join([generic_tbl_name, "id"]),
        col_author_id="_".join([Individual.__tablename__, "id"]),
        target_id=generic_id,
    )


def get_authored_object(
    c: Connection, generic_tbl_name: str, generic_id: str
) -> dict:
    tbl = generic_tbl_name
    idx = generic_id
    a = c.db.execute_returning_dicts(sql_get_detail(tbl, idx))[0]
    b = c.db.execute_returning_dicts(sql_get_authors(tbl, idx))[0]
    result = a | b
    return result



from sqlpyd import Connection

from corpus_x import Statute
from corpus_x.codifications import CodeRow, Codification
from corpus_x.inclusions import (
    CitationInOpinion,
    Inclusion,
    StatuteInOpinion,
    populate_db_with_inclusions,
)


def setup(c: Connection):
    """
    Assumes that (1) tables have been deleted with
    `corpus_x.utils.del_tables.delete_tables_with_prefix()`
    and (2) inclusion `.yaml` files previously been created with
    `corpus_x.inclusions.create_inclusion_files_from_db_opinions()`
    """
    # initialize tables
    Statute.make_tables(c)
    Inclusion.make_tables(c)
    Codification.make_tables(c)

    # The inclusion tables will be populated from inclusion .yaml files
    populate_db_with_inclusions(c)

    # The inclusion tables will now contain a reference to a statute but the statute doesn't exist yet
    StatuteInOpinion.add_statutes(c)
    StatuteInOpinion.update_statute_ids(c)

    # The inclusion tables will contain a reference to a decision; this needs to be updated with the id
    CitationInOpinion.update_decision_ids(c)

    # Since statutes already exist, can proceed to add codifications
    Codification.add_rows(c)
    for row in c.db[CodeRow.__tablename__].rows:
        CodeRow.set_update_units(c, row["id"])



from pathlib import Path

import yaml


def create_data_file(obj: dict, folder: Path, replace: bool = False) -> Path:
    """The `folder` represents where the `data.yaml` file should be created. This is typically the parent path of the main files used to generate the model. If `replace` is `True`, update the data file if it already exists."""
    if not folder.is_dir():
        raise Exception("Use folder to create `data.yaml`.")
    data_file = folder / "data.yaml"
    if data_file.exists():
        if replace:
            data_file.unlink()
            with open(data_file, "w") as f:
                yaml.safe_dump(obj, f)
    else:
        with open(data_file, "w") as f:
            yaml.safe_dump(obj, f)
    return data_file



from sqlpyd import Connection


def delete_tables_with_prefix(c: Connection, target_prefixes: list[str]):
    """Delete tables containing  `table_prefix-` found in  database connection `c`."""
    for p in target_prefixes:
        sql = f"""--sql
            select name
            from sqlite_schema
            where type='table' and name like '{p}%'
            order by name;
            """
        for row in c.db.execute_returning_dicts(sql):
            c.db.execute(
                f"""--sql
                drop table if exists {row['name']};
            """
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['corpus_x', 'corpus_x.utils']

package_data = \
{'': ['*'],
 'corpus_x': ['sql/analysis/*',
              'sql/base/*',
              'sql/codes/*',
              'sql/codes/events/*',
              'sql/decisions/*',
              'sql/decisions/inclusions/*',
              'sql/statutes/*',
              'sql/statutes/references/*']}

install_requires = \
['corpus-base>=0.1.0,<0.2.0', 'statute-trees>=0.0.18,<0.0.19']

setup_kwargs = {
    'name': 'corpus-x',
    'version': '0.0.1',
    'description': 'Add codification and statute tables to pre-existing corpus-base database.',
    'long_description': '# corpus-x\n\n## Concept\n\n[corpus-pax](https://github.com/justmars/corpus-pax) + [corpus-base](https://github.com/justmars/corpus-base) + [statute-trees](https://github.com/justmars/statute-trees) = converts raw `yaml`-based corpus repository to its database variant **corpus-x**; see [details](notebooks/setup.ipynb). After constructing all of the required tables, it becomes possible to [evaluate the raw data](docs/5-db-evaluate.md).\n\n## Flow\n\n### Local files\n\nDownload *.yaml files from repository:\n\n```mermaid\nflowchart LR\nrepo(((github/corpus))) --download---> local(local machine)\n```\n\n### Local database\n\nSetup local db:\n\n```mermaid\nflowchart LR\n  local(local corpus)--add corpus-pax tables--->db\n  local--add corpus-base tables-->db\n  local--format trees with statute-trees-->trees\n  trees(tree structures)--add corpus-x tables-->db[(sqlite.db)]\n```\n\n### Replicated database\n\nStore backup db on aws:\n\n```mermaid\nflowchart LR\n\n  db[(sqlite.db)]--litestream replicate-->aws\n  aws--litestream restore-->lawdata.xyz\n```\n\n## Mode\n\nOrder | Time | Instruction | Docs\n:--:|:--:|--:|:--\n0 | ~6sec (if with test data) | [corpus-pax](https://github.com/justmars/corpus-pax#read-me) pre-requiste before `corpus-base` can work. |[Setup](docs/1-setup.md)\n1 | ~20-40min | [corpus-base](https://github.com/justmars/corpus-base#read-me) pre-requiste before `corpus-x` can work. |[Setup](docs/1-setup.md)\n2 | ~120-130min | If inclusion files not yet created, run script to generate. |[Pre-inclusions](docs/2-pre-inclusions.md)\n3 | ~10min | Assuming inclusion files are already created, can populate the various tables under `corpus-x` | [Post-inclusions](docs/3-post-inclusions.md)\n4 | ~60min | Litestream output `x.db` on AWS bucket | [Replicated db](docs/4-aws-replicate.md)\n\n## Gotchas\n\nThe statutory event data contained in the `units` field does not yet contain the `statute_id`s. Note that, prior to database insertion, we only know the statute label but not the id. Once the statute has been inserted, we can now match the statute label to the id:\n\n```python\nfor row in c.db[CodeRow.__tablename__].rows:\n    obj = CodeRow.set_update_units(c, row["id"])\n```\n',
    'author': 'Marcelino G. Veloso III',
    'author_email': 'mars@veloso.one',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://lawdata.xyz',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '==3.11.0',
}


setup(**setup_kwargs)



from .__main__ import (
    build_x_tables,
    get_codification,
    get_decision,
    get_document,
    get_statute,
)
from .codifications import Codification
from .documents import Document
from .inclusions import CitationInOpinion, Inclusion, StatuteInOpinion
from .statutes import Statute



import json
import re
from datetime import datetime


def pluralize(number, singular="", plural="s"):
    """countr{{ num_countries|pluralize:("y","ies") }} See
    https://stackoverflow.com/questions/11714614/how-to-pluralize-a-name-in-a-
    template-with-jinja2/22336061#22336061."""
    if number == 1:
        return singular
    else:
        return plural


def get_year(value: str, short: bool = False, format: str = "%Y-%m-%d") -> str:
    """The date format is the default from sql result queries; if short is
    enabled only the last 2 digits of the year are included."""
    date_obj = datetime.strptime(value, format)
    year = date_obj.strftime("%Y")
    return year if not short else f"'{year[-2:]}"


def readable_date(value: str, format: str = "%Y-%m-%d") -> str:
    """The date format is the default from sql result queries; if short is
    enabled only the last 2 digits of the year are included."""
    date_obj = datetime.strptime(value, format)
    readable = date_obj.strftime("%B %-d, %Y")
    return readable


def make_unique_titles(value: str) -> str:
    """The titles may consist of duplicate text since the official title might
    also be the short title, etc."""
    texts = value.split("; ")
    uniques = set(texts)
    return "; ".join(uniques)


def extract_units(raw_units: str) -> list[dict]:
    """Removes the initial `1.` material path which maps to the title of the
    tree; the units start with the content."""
    units = json.loads(raw_units)
    if units:
        return units[0]["units"]
    return []


def valid_mp(candidate_unit_id: str) -> str:
    """Each unit id is material path following the format: 1.1.1.1.

    etc. where all root nodes start with 1. and all branches / leaves
    terminate in the same fashion, e.g. 1.2.13.16.

    The root's identifier `materialized_path` is always 1. and the value of the specific branch can be determined through the addition of other identifier's to the root. e.g. '1.1.1.2' of 'mv-2022-ra-386-v1' will yield all material nodes of the tree from the root to the specified `mp` branch.
    """
    mp_pattern = re.compile(r"(\d+\.)+")
    if mp_pattern.fullmatch(candidate_unit_id):
        return candidate_unit_id
    raise Exception(f"Bad {candidate_unit_id=}")


def shorten_item(text: str):
    """Replace the item's long form keyword with its shorthand format, if
    possible."""
    if "Republic Act" in text:
        return re.sub(r"Republic\s+Act\s+No\.", "RA", text)
    if "Presidential Decree" in text:
        return re.sub(r"Presidential\s+Decree\s+No\.", "PD", text)
    if "Batas Pambansa" in text:
        return re.sub(r"Batas\s+Pambansa\s+Blg\.", "BP", text)
    if "Executive Order" in text:
        return re.sub(r"Executive\s+Order\s+No\.", "EO", text)
    if "Commonwealth Act" in text:
        return re.sub(r"Commonwealth\s+Act\s+No\.", "CA", text)
    if "Administrative Matter" in text:
        return re.sub(r"Administrative\s+Matter\s+No\.", "AM", text)
    if "Bar Matter" in text:
        return re.sub(r"Bar\s+Matter\s+No\.", "BM", text)
    if "Administrative Circular" in text:
        return re.sub(r"Administrative\s+Circular\s+No\.", "AC", text)
    if "Paragraph" in text:
        return re.sub(r"Paragraph", "Par.", text)
    if "Article" in text:
        return re.sub(r"Article", "Art.", text)
    if "Section" in text:
        return re.sub(r"Section", "Sec.", text)
    if "Book" in text:
        return re.sub(r"Book", "Bk.", text)
    if "Chapter" in text:
        return re.sub(r"Chapter", "Ch.", text)
    if "Sub-Container" in text:
        return re.sub(r"Sub-Container(\s+\d+)?", "", text)
    if "Container" in text:
        return re.sub(r"Container(\s+\d+)?", "", text)
    return text
